/********************************************************************************

 Programmed by SEUNG-JO BAE 

  - Goal: Reduction & Prefix Sum (Inclusive & Exclusive)

  - Algorithm: Kurmar's Book Program 3.6 & 3.7

  - Input: Integer number

  - Output: Integer number

  - Version 1.0 (2/13/95)

  - Communication: CMMD Swap

  - Note: 


********************************************************************************/

#include <cm/cmmd.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <cmsys/cm_fcntl.h>  /* fcntl( fileno(stdout), F_SETFL, O_APPEND); */

#include "global0.h"

#include "function0.h"


void initialize()
{
   Boolean error;
   int i;

   CMMD_disable_safety();
   CMMD_fset_io_mode(stdin, CMMD_sync_bc);
   CMMD_fset_io_mode(stdout, CMMD_sync_bc);
   CMMD_fset_io_mode(stderr, CMMD_independent);

   ori_nproc = CMMD_partition_size();
   do {
      printf("\nEnter number of processors(Max=%d): ", ori_nproc);
      scanf("%d", &nproc);
      error = ( !Blog2(nproc) || nproc<2 || nproc>ori_nproc || nproc>N );
      if (error)
	 printf("\n??? Improper input. Try again!\n");
    } while (error);
    i = CMMD_reset_partition_size(nproc);
    if (i != nproc) 
       CMMD_error("\nERROR: CMMD Partition failed!\n");
   
   my_pnum  = CMMD_self_address(); 
 }

int reduction_dim0(my_num, d, result)
  int my_num, d, *result;
{

  int in_mesg;
  int partner;
  int i;


  CMMD_sync_with_nodes();

  *result = my_num;

  for (i=0; i<d; ++i) {
    partner = my_pnum ^ pow2(i);
    CMMD_swap(partner, 
	      &in_mesg, 4,   /* receive */
	      result, 4);    /* send */
    *result += in_mesg;
  }

  CMMD_sync_with_nodes();
}


int in_prefix_dim0(my_num, d, result)   /* Inclusive sum */
  int my_num, d, *result;
{

  int out_mesg, in_mesg;
  int partner;
  int i;


  CMMD_sync_with_nodes();

  *result = my_num;
  out_mesg = my_num;

  for (i=0; i<d; ++i) {
    partner = my_pnum ^ pow2(i);
    CMMD_swap(partner, 
	      &in_mesg, 4,      /* receive */
	      &out_mesg, 4);    /* send */
    out_mesg += in_mesg;
    if (partner < my_pnum)
      *result += in_mesg;
  }

  CMMD_sync_with_nodes();
}

int ex_prefix_dim0(my_num, d, result)   /* Inclusive sum */
  int my_num, d, *result;
{

  int out_mesg, in_mesg;
  int partner;
  int i;


  CMMD_sync_with_nodes();

  *result = 0;
  out_mesg = my_num;

  for (i=0; i<d; ++i) {
    partner = my_pnum ^ pow2(i);
    CMMD_swap(partner, 
	      &in_mesg, 4,       /* receive */
	      &out_mesg, 4);     /* send */
    out_mesg += in_mesg;
    if (partner < my_pnum)
      *result += in_mesg;
  }

  CMMD_sync_with_nodes();
}

void main()
{

  int my_num, d, result;

  initialize();

  my_num = 1;
  d = Ilog2(nproc);

  reduction_dim0(my_num, d, &result);

  fcntl( fileno(stdout), F_SETFL, O_APPEND);
  CMMD_fset_io_mode(stdout, CMMD_independent);

  printf("\n [pn=%2d] in=%d, out=%d \n", my_pnum, my_num, result);
  fflush(stdout);

  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n------------------------------\n");

  in_prefix_dim0(my_num, d, &result);

  fcntl( fileno(stdout), F_SETFL, O_APPEND);
  CMMD_fset_io_mode(stdout, CMMD_independent);

  printf("\n [pn=%2d] in=%d, out=%d \n", my_pnum, my_num, result);
  fflush(stdout);

  CMMD_fset_io_mode(stdout, CMMD_sync_bc);
  printf("\n------------------------------\n");

  ex_prefix_dim0(my_num, d, &result);

  fcntl( fileno(stdout), F_SETFL, O_APPEND);
  CMMD_fset_io_mode(stdout, CMMD_independent);

  printf("\n [pn=%2d] in=%d, out=%d \n", my_pnum, my_num, result);
  fflush(stdout);

}
